﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using ICallisto;

namespace callisto
{
    class CallistoPluginHost : IHost
    {
        public CallistoPluginHost(String datapath)
        {
            this.datapath = datapath;
        }

        private CallistoPluginScript script = new CallistoPluginScript();
        public IScript Script { get { return this.script; } }

        private CallistoPluginRoom room = new CallistoPluginRoom();
        public IRoom Room { get { return this.room; } }

        private CallistoPluginLinker linker = new CallistoPluginLinker();
        public ILinker Linker { get { return this.linker; } }

        private CallistoPluginStats stats = new CallistoPluginStats();
        public IStats Stats { get { return this.stats; } }

        private CallistoPluginHashlink hashlink = new CallistoPluginHashlink();
        public IHashlink Hashlink { get { return this.hashlink; } }

        private CallistoPluginCompress compress = new CallistoPluginCompress();
        public ICompress Zip { get { return this.compress; } }

        private String datapath;
        public String DataPath { get { return this.datapath; } }

        public IPEndPoint[] ServerNodes
        {
            get
            {
                List<IPEndPoint> list = new List<IPEndPoint>();
                UdpServerPool.Pool.ForEach(x => list.Add(new IPEndPoint(x.ip, x.port)));
                return list.ToArray();
            }
        }

        public IUser FirstLinked
        {
            get
            {
                if (!CallistoServer.Linker.IsLinking)
                    return null;

                if (CallistoServer.Linker.linked_users.Count == 0)
                    return null;

                return CallistoServer.Linker.linked_users[0];
            }
        }

        public IUser FirstUser
        {
            get
            {
                foreach (UserObject u in UserPool.Pool)
                    if (u.Connected)
                        if (u.logged_in)
                            return u;

                foreach (WebUser u in WebPool.Pool)
                    if (u.InUse)
                        if (u.LoggedIn)
                            return u;

                return null;
            }
        }

        public IBan FirstBanned
        {
            get { return BanManager.GetFirstBanned; }
        }

        public void ClearBans()
        {
            BanManager.Clear();
        }

        public IUser User(ushort id)
        {
            foreach (UserObject u in UserPool.Pool)
                if (u.Connected)
                    if (u.logged_in)
                        if (u.id == id)
                            return u;

            foreach (WebUser u in WebPool.Pool)
                if (u.InUse)
                    if (u.LoggedIn)
                        if (u.Id == id)
                            return u;

            return null;
        }

        public IUser User(String name)
        {
            foreach (UserObject u in UserPool.Pool)
                if (u.Connected)
                    if (u.logged_in)
                        if (u.name.StartsWith(name))
                            return u;

            foreach (WebUser u in WebPool.Pool)
                if (u.InUse)
                    if (u.LoggedIn)
                        if (u.Name.StartsWith(name))
                            return u;

            if (CallistoServer.Linker.IsLinking)
            {
                IUser u = CallistoServer.Linker.linked_users.Find(l => l.name.StartsWith(name));

                if (u != null)
                    return u;
            }

            return null; // nobody home
        }

        public void SendURL(String address, object text)
        {
            if (text != null)
                UserPool.BroadcastToAll(ChatPacketHelpers.Url(address, text.ToString()));
        }

        public void PrintToAdmins(object text)
        {
            if (text != null)
                UserPool.BroadcastToAdmins(ChatPacketHelpers.NoSuch(text.ToString()));
        }

        public void PrintToAdmins(object text, bool link)
        {
            if (text == null)
                return;
            
            UserPool.BroadcastToAdmins(ChatPacketHelpers.NoSuch(text.ToString()));

            if (link)
                if (CallistoServer.Linker.IsLinking)
                    CallistoServer.Linker.SendPacket(LinkPacketHelpers.AdminMessage(text.ToString()));
        }

        public void PrintToLink(object text)
        {
            if (text == null)
                return;

            if (CallistoServer.Linker.IsLinking)
                CallistoServer.Linker.SendPacket(LinkPacketHelpers.Broadcast(ChatPacketHelpers.NoSuch(text.ToString())));
        }

        public void Print(ushort vroom, object text)
        {
            if (text == null)
                return;

            UserPool.BroadcastToVroom(vroom, ChatPacketHelpers.NoSuch(text.ToString()));
            WebPool.BroadcastPrint(vroom, text.ToString());
        }

        public void Print(ushort vroom, object text, bool link)
        {
            if (text == null)
                return;

            UserPool.BroadcastToVroom(vroom, ChatPacketHelpers.NoSuch(text.ToString()));
            WebPool.BroadcastPrint(vroom, text.ToString());

            if (link)
                if (CallistoServer.Linker.IsLinking)
                    CallistoServer.Linker.SendPacket(LinkPacketHelpers.Broadcast(ChatPacketHelpers.NoSuch(text.ToString())));
        }

        public void Print(IUser user, object text)
        {
            if (user != null)
                user.SendPrint(text.ToString());
        }

        public void Print(object text)
        {
            if (text == null)
                return;

            UserPool.BroadcastToAll(ChatPacketHelpers.NoSuch(text.ToString()));
            WebPool.BroadcastPrint(text.ToString());
        }

        public void Print(object text, bool link)
        {
            if (text == null)
                return;

            UserPool.BroadcastToAll(ChatPacketHelpers.NoSuch(text.ToString()));
            WebPool.BroadcastPrint(text.ToString());

            if (link)
                if (CallistoServer.Linker.IsLinking)
                    CallistoServer.Linker.SendPacket(LinkPacketHelpers.Broadcast(ChatPacketHelpers.NoSuch(text.ToString())));
        }

        public void SendPM(IUser target, String sender, String text)
        {
            if (target == null)
                return;

            if (target is UserObject)
            {
                UserObject u = (UserObject)target;
                u.SendPluginPM(sender, text);
            }
            else if (target is LinkedUserObject)
            {
                if (CallistoServer.Linker.IsLinking)
                    CallistoServer.Linker.SendPacket(LinkPacketHelpers.ForcePM(target, sender, text));                    
            }
        }

        public void SendText(IUser target, String sender, String text)
        {
            if (target == null)
                return;

            if (target is UserObject)
            {
                UserObject u = (UserObject)target;
                u.SendPacket(ChatPacketHelpers.Public(sender, text));
            }
            else if (target is WebUser)
            {
                WebUser u = (WebUser)target;
                u.QueuePacket(WebPackets.PublicTo(u, sender, text));
            }
            else if (target is LinkedUserObject)
            {
                if (CallistoServer.Linker.IsLinking)
                    CallistoServer.Linker.SendPacket(LinkPacketHelpers.RelayToUser(target.Name, ChatPacketHelpers.Public(sender, text)));
            }
        }

        public void SendEmote(IUser target, String sender, String text)
        {
            if (target == null)
                return;

            if (target is UserObject)
            {
                UserObject u = (UserObject)target;
                u.SendPacket(ChatPacketHelpers.Emote(sender, text));
            }
            else if (target is WebUser)
            {
                WebUser u = (WebUser)target;
                u.QueuePacket(WebPackets.EmoteTo(u, sender, text));
            }
            else if (target is LinkedUserObject)
            {
                if (CallistoServer.Linker.IsLinking)
                    CallistoServer.Linker.SendPacket(LinkPacketHelpers.RelayToUser(target.Name, ChatPacketHelpers.Emote(sender, text)));
            }
        }
    }
}
