﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Net.Sockets;
using ICallisto;

namespace callisto
{
    class LinkClient : ILink
    {
        private List<byte> data_in = new List<byte>();
        private Queue<byte[]> data_out = new Queue<byte[]>();
        private List<AresPacket> formatted_packets = new List<AresPacket>();

        internal List<LinkedUserObject> linked_users = new List<LinkedUserObject>();
        internal int receive_fails = 0;

        internal String hashlink = String.Empty;
        internal String room_name = String.Empty;
        internal LinkMode mode = LinkMode.Server;
        internal Socket socket = null;
        internal IPAddress target_ip = IPAddress.Any;
        internal ushort target_port = 0;
        internal bool connected = false;
        internal bool logged_in = false;
        internal uint session_spawn_time = 0;
        internal uint last_update_sent_time = 0;
        internal uint last_update_received_time = 0;
        internal ushort protocol = 0;

        internal bool IsActive { get { return this.socket != null; } }
        internal bool IsLinking { get { return this.socket != null && this.logged_in; } }


        public String Hashlink { get { return this.hashlink; } }

        public ushort UserCount { get { return (ushort)this.linked_users.Count; } }

        public String Name { get { return this.room_name; } }

        public IPAddress ExternalIP { get { return this.target_ip; } }

        public ushort Port { get { return this.target_port; } }

        public LinkMode Mode { get { return this.mode; } }

        public LinkClient() { }

        /// <summary>we are the server</summary>
        public LinkClient(Socket socket, uint session_spawn_time) // we are the server
        {
            this.mode = LinkMode.Server;
            this.session_spawn_time = session_spawn_time;
            this.last_update_sent_time = session_spawn_time;
            this.last_update_received_time = session_spawn_time;
            this.socket = socket;
            this.socket.Blocking = false;
            this.target_ip = ((IPEndPoint)(socket.RemoteEndPoint)).Address;
            this.connected = true;
            this.logged_in = false;
            this.receive_fails = 0;
            this.data_in.Clear();
            this.data_out.Clear();
            this.formatted_packets.Clear();
        }

        /// <summary>we are the client</summary>
        public void Connect(String hashlink, uint session_spawn_time)
        {
            if (CallistoServer.Linker.IsActive || !Settings.RoomLinking) // go away
            {
                ServerEvents.OnLinkError(this, LinkError.LinkingDisabled);
                return;
            }

            ChannelObject chan = HashlinkHelpers.DecodeHashlink(hashlink);

            if (chan == null)
            {
                ServerEvents.OnLinkError(this, LinkError.InvalidHashlink);
                return;
            }

            if (Helpers.IsLocalHost(chan.external_ip) && ((ushort)Settings.ServerPort) == chan.port)
            {
                ServerEvents.OnLinkError(this, LinkError.LinkToSelf);
                return;
            }

            this.mode = LinkMode.Client;
            this.session_spawn_time = session_spawn_time;
            this.last_update_sent_time = session_spawn_time;
            this.last_update_received_time = session_spawn_time;
            this.socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            this.socket.Blocking = false;
            this.target_ip = chan.external_ip;
            this.target_port = chan.port;
            this.connected = false;
            this.logged_in = false;
            this.receive_fails = 0;
            this.data_in.Clear();
            this.data_out.Clear();
            this.formatted_packets.Clear();
            this.linked_users.Clear();
            this.SendPacket(LinkPacketHelpers.RequestLinkPacket());

            try // would block
            {
                this.socket.Connect(new IPEndPoint(chan.external_ip, chan.port));
            }
            catch { }
        }

        internal bool ClientConnected(uint time)
        {
            if (!this.connected)
            {
                if (time > (this.session_spawn_time + 15)) // unable to connect
                {
                    ServerEvents.OnLinkError(this, LinkError.PingTimeout);
                    this.Disconnect();
                    return false;
                }

                if (this.socket.Poll(0, SelectMode.SelectWrite)) // connected :-)
                {
                    this.connected = true;
                    this.session_spawn_time = time; // reset spawn time for next check
                    return true;
                }

                return false;
            }

            return true;
        }

        internal void SendAndDisconnect()
        {
            this.DispatchOutbound();

            try { this.socket.Disconnect(false); }
            catch { }
            try { this.socket.Close(); }
            catch { }

            this.socket = null;
            this.connected = false;
            this.linked_users.Clear();
        }

        internal void Disconnect()
        {
            try { this.socket.Disconnect(false); }
            catch { }
            try { this.socket.Close(); }
            catch { }

            this.socket = null;
            this.connected = false;
            this.linked_users.Clear();
        }

        internal void PartAndDisconnect()
        {
            if (this.linked_users.Count > 0)
            {
                foreach (LinkedUserObject l in this.linked_users)
                {
                    if (l.visible)
                    {
                        UserPool.BroadcastToVroom(0, LinkPacketHelpers.Part(l));
                        ServerEvents.OnPart(l);
                    }
                }
            }

            try { this.socket.Disconnect(false); }
            catch { }
            try { this.socket.Close(); }
            catch { }

            if (this.IsLinking)
                ServerEvents.OnLinkPart(this);

            this.socket = null;
            this.connected = false;
            this.linked_users.Clear();
        }

        internal void SendPacket(byte[] data)
        {
            this.data_out.Enqueue(data);
        }

        internal void DispatchOutbound()
        {
            while (this.data_out.Count > 0)
            {
                try
                {
                    byte[] buf = this.data_out.Peek();
                    this.socket.Send(buf);
                    CallistoPluginStats.data_sent += (ulong)this.data_out.Dequeue().Length;
                }
                catch { return; }
            }
        }

        internal void BuildHashlink()
        {
            ChannelObject chan = new ChannelObject();
            chan.name = this.room_name;
            chan.topic = String.Empty;
            chan.external_ip = this.target_ip;
            chan.local_ip = IPAddress.Any;
            chan.port = this.target_port;
            this.hashlink = HashlinkHelpers.EncodeHashlink(chan);
        }

        internal void ReceiveInbound()
        {
            int receive_count = 0;
            byte[] buffer1 = new byte[8192];
            SocketError e = SocketError.Success;

            try
            {
                int avail = this.socket.Available;

                if (avail > 8192)
                    avail = 8192;

                receive_count = this.socket.Receive(buffer1, 0, avail, SocketFlags.None, out e);
            }
            catch { } // would block

            if (receive_count == 0)
            {
                if (e == SocketError.WouldBlock)
                    this.receive_fails = 0;
                else
                    this.receive_fails++;
            }
            else
            {
                CallistoPluginStats.data_received += (ulong)receive_count;
                byte[] buffer2 = new byte[receive_count];
                Array.Copy(buffer1, buffer2, receive_count);
                this.data_in.AddRange(buffer2);
                this.receive_fails = 0;
            }

            while (this.data_in.Count >= 3)
            {
                AresPacket packet = new AresPacket();
                packet.Size = BitConverter.ToUInt16(this.data_in.ToArray(), 0);
                packet.Msg = (ProtoMessage)this.data_in[2];

                if (this.data_in.Count >= (packet.Size + 3))
                {
                    this.data_in.RemoveRange(0, 3); // no longer need the header
                    packet.Packet = this.data_in.GetRange(0, packet.Size).ToArray();
                    this.data_in.RemoveRange(0, packet.Size);

                    if (packet.Msg == ProtoMessage.MSG_CHAT_CLIENTCOMPRESSED) // convert into packet array
                    {
                        try
                        {
                            AresPacket[] zip_packets = Helpers.GetPacketsFromZip(packet.Packet);
                            this.formatted_packets.AddRange(zip_packets);
                        }
                        catch { }
                    }
                    else { this.formatted_packets.Add(packet); } // add to the heap
                }
                else { break; }
            }
        }

        internal AresPacket NextPacket
        {
            get
            {
                if (this.formatted_packets.Count == 0)
                    return null;

                AresPacket packet = this.formatted_packets[0];
                this.formatted_packets.RemoveAt(0);
                return packet;
            }
        }


    }
}
