﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Net;
using System.IO;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using Jurassic;
using Jurassic.Library;
using ICallisto;

namespace callisto.Scripting.Objects
{
    class Avatar : ClrFunction
    {
        public Avatar(ScriptEngine engine)
            : base(engine.Function.InstancePrototype, "Avatar", new AvatarInstance(engine.Object.InstancePrototype))
        {

        }

        [JSConstructorFunction]
        public AvatarInstance Construct()
        {
            return new AvatarInstance(this.InstancePrototype);
        }
    }

    public class AvatarInstance : ObjectInstance
    {
        private bool busy = false;

        public AvatarInstance(ObjectInstance prototype)
            : base(prototype)
        {
            this.PopulateFunctions();
            this.DefineProperty("download", new PropertyDescriptor(this["download"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("load", new PropertyDescriptor(this["load"], PropertyAttributes.Sealed), throwOnError: true);
        }

        [JSProperty(Name = "src")]
        public String Source { get; set; }

        [JSProperty(Name = "oncomplete")]
        public UserDefinedFunction Callback { get; set; }

        [JSFunction(Name = "download")]
        public bool Download(object a)
        {
            if (this.busy)
                return false;

            Thread thread = new Thread(new ThreadStart(() =>
            {
                this.busy = true;
                String arg = String.Empty;

                if (!(a is Undefined) && a != null)
                    arg = a.ToString();

                AvatarImage result = new AvatarImage(this.Engine.Object.InstancePrototype)
                {
                    Callback = this.Callback,
                    Data = null,
                    ScriptName = this.Engine.ScriptName,
                    Arg = arg
                };

                try
                {
                    WebRequest request = WebRequest.Create(this.Source);
                    WebResponse response = request.GetResponse();
                    List<byte> bytes_in = new List<byte>();
                    int received = 0;
                    byte[] buf = new byte[1024];

                    using (Stream stream = response.GetResponseStream())
                        while ((received = stream.Read(buf, 0, 1024)) > 0)
                            bytes_in.AddRange(buf.Take(received));

                    response.Close();

                    using (Bitmap avatar_raw = new Bitmap(new MemoryStream(bytes_in.ToArray())))
                    {
                        using (Bitmap avatar_sized = new Bitmap(48, 48))
                        {
                            using (Graphics g = Graphics.FromImage(avatar_sized))
                            {
                                g.InterpolationMode = InterpolationMode.HighQualityBicubic;
                                g.DrawImage(avatar_raw, new RectangleF(0, 0, 48, 48));

                                using (MemoryStream ms = new MemoryStream())
                                {
                                    avatar_sized.Save(ms, ImageFormat.Jpeg);
                                    result.Data = ms.ToArray();
                                    bytes_in.Clear();
                                }
                            }
                        }
                    }
                }
                catch { }

                SyncList.AddItem(result);
                this.busy = false;
            }));

            thread.Start();

            return true;
        }

        private String[] bad_chars = new String[]
        {
            ".",
            "/",
            "\\",
            " ",
        };

        [JSFunction(Name = "load")]
        public AvatarImage Load(object a)
        {
            AvatarImage av = new AvatarImage(this.Engine.Object.InstancePrototype);

            if (a is String || a is ConcatenatedString)
            {
                String filename = a.ToString();

                if (filename.Length > 1)
                    if (bad_chars.Count<String>(x => filename.Contains(x)) == 0)
                    {
                        String path = Settings.Path + "Scripts\\" + this.Engine.ScriptName + "\\data\\" + filename;

                        try
                        {
                            if (File.Exists(path))
                            {
                                byte[] data = File.ReadAllBytes(path);

                                using (Bitmap avatar_raw = new Bitmap(new MemoryStream(data)))
                                {
                                    using (Bitmap avatar_sized = new Bitmap(48, 48))
                                    {
                                        using (Graphics g = Graphics.FromImage(avatar_sized))
                                        {
                                            g.InterpolationMode = InterpolationMode.HighQualityBicubic;
                                            g.DrawImage(avatar_raw, new RectangleF(0, 0, 48, 48));

                                            using (MemoryStream ms = new MemoryStream())
                                            {
                                                avatar_sized.Save(ms, ImageFormat.Jpeg);
                                                av.Data = ms.ToArray();
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        catch { }
                    }
            }

            return av;
        }
    }

    public class AvatarImage : ObjectInstance, ISyncListItem
    {
        public AvatarImage(ObjectInstance prototype)
            : base(prototype)
        {
            this.PopulateFunctions();
            this.DefineProperty("download", new PropertyDescriptor(this["download"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("save", new PropertyDescriptor(this["save"], PropertyAttributes.Sealed), throwOnError: true);
        }

        public byte[] Data { get; set; }
        public UserDefinedFunction Callback { get; set; }
        public String ScriptName { get; set; }
        public String Arg { get; set; }

        [JSProperty(Name = "arg")]
        public String GetArgument
        {
            get { return this.Arg; }
            set { }
        }

        [JSProperty(Name = "exists")]
        public bool DoesExist
        {
            get { return this.Data != null; }
            set { }
        }

        [JSFunction(Name = "toScribble")]
        public ScribbleResult ToScribble()
        {
            if (this.Data != null)
            {
                ScribbleResult s = new ScribbleResult(this.Engine.Object.InstancePrototype);
                s.FromAvatar(this.Data);
                return s;
            }

            return null;
        }

        private String[] bad_chars = new String[]
        {
            ".",
            "/",
            "\\",
            " ",
        };

        [JSFunction(Name = "save")]
        public bool Save(object a)
        {
            if (this.Data == null)
                return false;

            if (a is String || a is ConcatenatedString)
            {
                String filename = a.ToString();

                if (filename.Length > 1)
                    if (bad_chars.Count<String>(x => filename.Contains(x)) == 0)
                    {
                        String path = Settings.Path + "Scripts\\" + this.Engine.ScriptName + "\\data";

                        try
                        {
                            if (!Directory.Exists(path))
                                Directory.CreateDirectory(path);

                            path += "\\" + filename;

                            File.WriteAllBytes(path, this.Data);
                            return true;
                        }
                        catch { }
                    }
            }

            return false;
        }

        public void FromScribble(byte[] data)
        {
            try
            {
                using (Bitmap avatar_raw = new Bitmap(new MemoryStream(data)))
                {
                    using (Bitmap avatar_sized = new Bitmap(48, 48))
                    {
                        using (Graphics g = Graphics.FromImage(avatar_sized))
                        {
                            g.InterpolationMode = InterpolationMode.HighQualityBicubic;
                            g.DrawImage(avatar_raw, new RectangleF(0, 0, 48, 48));

                            using (MemoryStream ms = new MemoryStream())
                            {
                                avatar_sized.Save(ms, ImageFormat.Jpeg);
                                this.Data = ms.ToArray();
                            }
                        }
                    }
                }
            }
            catch { }
        }

        public override string ToString()
        {
            return "[object AvatarImage]";
        }
    }
}
