﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Net;
using System.IO;
using Jurassic;
using Jurassic.Library;

namespace callisto.Scripting.Objects
{
    class HttpRequest : ClrFunction
    {
        public HttpRequest(ScriptEngine engine)
            : base(engine.Function.InstancePrototype, "HttpRequest", new HttpRequestInstance(engine.Object.InstancePrototype))
        {

        }

        [JSConstructorFunction]
        public HttpRequestInstance Construct()
        {
            return new HttpRequestInstance(this.InstancePrototype);
        }
    }

    public class HttpRequestInstance : ObjectInstance
    {
        private bool busy = false;
        private String _method = "get";
        private List<HttpHeaderItem> _headers = new List<HttpHeaderItem>();

        public HttpRequestInstance(ObjectInstance prototype)
            : base(prototype)
        {
            this.PopulateFunctions();
            this.Agent = String.Empty;
            this.DefineProperty("download", new PropertyDescriptor(this["download"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("header", new PropertyDescriptor(this["header"], PropertyAttributes.Sealed), throwOnError: true);
        }

        [JSProperty(Name = "method")]
        public String Method
        {
            get { return this._method; }
            set
            {
                String str = value.ToUpper();

                if (str == "POST")
                    this._method = "post";
                else if (str == "GET")
                    this._method = "get";
            }
        }

        [JSProperty(Name = "params")]
        public String Params { get; set; }

        [JSProperty(Name = "src")]
        public String Source { get; set; }

        [JSProperty(Name = "userAgent")]
        public String Agent { get; set; }

        [JSProperty(Name = "oncomplete")]
        public UserDefinedFunction Callback { get; set; }

        [JSProperty(Name = "utf")]
        public bool UTF { get; set; }

        [JSProperty(Name = "host")]
        public String Host { get; set; }

        [JSFunction(Name = "header")]
        public bool Header(object key, object value)
        {
            if (key == null || value == null)
                return false;

            String k = key.ToString();
            String v = value.ToString();

            this._headers.RemoveAll(x => x.Key == k);
            this._headers.Add(new HttpHeaderItem { Key = k, Value = v });

            return true;
        }

        [JSFunction(Name = "download")]
        public bool Download(object a)
        {
            if (this.busy)
                return false;

            Thread thread = new Thread(new ThreadStart(() =>
            {
                this.busy = true;
                bool use_utf = this.UTF;
                String arg = String.Empty;

                if (!(a is Undefined) && a != null)
                    arg = a.ToString();

                HttpRequestResult result = new HttpRequestResult(this.Engine.Object.InstancePrototype)
                {
                    Callback = this.Callback,
                    Data = null,
                    ScriptName = this.Engine.ScriptName,
                    Arg = arg
                };

                try
                {
                    HttpWebRequest request = (HttpWebRequest)WebRequest.Create(this.Source);
                    request.UserAgent = this.Agent;
                    request.Method = this._method;

                    if (!String.IsNullOrEmpty(this.Host))
                        request.Host = this.Host;

                    foreach (HttpHeaderItem h in this._headers)
                        request.Headers.Add(h.Key, h.Value);

                    if (!String.IsNullOrEmpty(this.Params) && this._method == "post")
                    {
                        byte[] p = this.UTF ? Encoding.UTF8.GetBytes(this.Params) : Encoding.Default.GetBytes(this.Params);
                        request.ContentType = "application/x-www-form-urlencoded";
                        request.ContentLength = p.Length;

                        using (Stream stream = request.GetRequestStream())
                            stream.Write(p, 0, p.Length);
                    }

                    WebResponse response = request.GetResponse();

                    List<byte> bytes_in = new List<byte>();
                    byte[] buf = new byte[1024];
                    int rec = 0;

                    using (Stream stream = response.GetResponseStream())
                        while ((rec = stream.Read(buf, 0, 1024)) > 0)
                            bytes_in.AddRange(buf.Take(rec));

                    response.Close();
                    result.Data = use_utf ? Encoding.UTF8.GetString(bytes_in.ToArray()) : Encoding.Default.GetString(bytes_in.ToArray());
                }
                catch { }

                SyncList.AddItem(result);
                this.busy = false;
            }));

            thread.Start();

            return true;
        }
    }

    class HttpRequestResult : ObjectInstance, ISyncListItem
    {
        public HttpRequestResult(ObjectInstance prototype)
            : base(prototype)
        {
            this.PopulateFunctions();
            this.Data = String.Empty;
        }

        public String Data { get; set; }
        public UserDefinedFunction Callback { get; set; }
        public String ScriptName { get; set; }
        public String Arg { get; set; }

        [JSProperty(Name = "arg")]
        public String GetArgument
        {
            get { return this.Arg; }
            set { }
        }

        [JSProperty(Name = "page")]
        public String Page
        {
            get { return this.Data; }
            set { }
        }

        public override string ToString()
        {
            return "[object HttpResult]";
        }
    }

    class HttpHeaderItem
    {
        public String Key { get; set; }
        public String Value { get; set; }
    }
}
