﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Net;
using System.IO;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using Jurassic;
using Jurassic.Library;
using ICallisto;

namespace callisto.Scripting.Objects
{
    class Scribble : ClrFunction
    {
        public Scribble(ScriptEngine engine)
            : base(engine.Function.InstancePrototype, "Scribble", new ScribbleInstance(engine.Object.InstancePrototype))
        {

        }

        [JSConstructorFunction]
        public ScribbleInstance Construct()
        {
            return new ScribbleInstance(this.InstancePrototype);
        }
    }

    public class ScribbleInstance : ObjectInstance
    {
        private bool busy = false;

        public ScribbleInstance(ObjectInstance prototype)
            : base(prototype)
        {
            this.PopulateFunctions();
            this.DefineProperty("download", new PropertyDescriptor(this["download"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("load", new PropertyDescriptor(this["load"], PropertyAttributes.Sealed), throwOnError: true);
        }

        [JSProperty(Name = "src")]
        public String Source { get; set; }

        [JSProperty(Name = "oncomplete")]
        public UserDefinedFunction Callback { get; set; }

        [JSFunction(Name = "download")]
        public bool Download(object a)
        {
            if (this.busy)
                return false;

            Thread thread = new Thread(new ThreadStart(() =>
            {
                this.busy = true;
                String arg = String.Empty;

                if (!(a is Undefined) && a != null)
                    arg = a.ToString();

                ScribbleResult result = new ScribbleResult(this.Engine.Object.InstancePrototype)
                {
                    Callback = this.Callback,
                    Data = null,
                    ScriptName = this.Engine.ScriptName,
                    Arg = arg
                };

                try
                {
                    WebRequest request = WebRequest.Create(this.Source);
                    WebResponse response = request.GetResponse();
                    List<byte> bytes_in = new List<byte>();
                    int received = 0;
                    byte[] buf = new byte[1024];

                    using (Stream stream = response.GetResponseStream())
                        while ((received = stream.Read(buf, 0, 1024)) > 0)
                            bytes_in.AddRange(buf.Take(received));

                    response.Close();

                    using (Bitmap avatar_raw = new Bitmap(new MemoryStream(bytes_in.ToArray())))
                    {
                        int img_x = avatar_raw.Width;
                        int img_y = avatar_raw.Height;

                        if (img_x > 384)
                        {
                            img_x = 384;
                            img_y = avatar_raw.Height - (int)Math.Floor(Math.Floor((double)avatar_raw.Height / 100) * Math.Floor(((double)(avatar_raw.Width - 384) / avatar_raw.Width) * 100));
                        }

                        if (img_y > 384)
                        {
                            img_x -= (int)Math.Floor(Math.Floor((double)img_x / 100) * Math.Floor(((double)(img_y - 384) / img_y) * 100));
                            img_y = 384;
                        }

                        using (Bitmap avatar_sized = new Bitmap(img_x, img_y))
                        {
                            using (Graphics g = Graphics.FromImage(avatar_sized))
                            {
                                g.InterpolationMode = InterpolationMode.HighQualityBicubic;
                                g.DrawImage(avatar_raw, new RectangleF(0, 0, img_x, img_y));

                                using (MemoryStream ms = new MemoryStream())
                                {
                                    avatar_sized.Save(ms, ImageFormat.Jpeg);
                                    result.Data = Zip.Compress(ms.ToArray());
                                    bytes_in.Clear();
                                }
                            }
                        }
                    }
                }
                catch { }

                SyncList.AddItem(result);
                this.busy = false;
            }));

            thread.Start();

            return true;
        }

        private String[] bad_chars = new String[]
        {
            ".",
            "/",
            "\\",
            " ",
        };

        [JSFunction(Name = "load")]
        public ScribbleResult Load(object a)
        {
            ScribbleResult scr = new ScribbleResult(this.Engine.Object.InstancePrototype);

            if (a is String || a is ConcatenatedString)
            {
                String filename = a.ToString();

                if (filename.Length > 1)
                    if (bad_chars.Count<String>(x => filename.Contains(x)) == 0)
                    {
                        String path = Settings.Path + "Scripts\\" + this.Engine.ScriptName + "\\data\\" + filename;

                        try
                        {
                            if (File.Exists(path))
                            {
                                byte[] data = File.ReadAllBytes(path);

                                try
                                {
                                    using (Bitmap avatar_raw = new Bitmap(new MemoryStream(data)))
                                    {
                                        int img_x = avatar_raw.Width;
                                        int img_y = avatar_raw.Height;

                                        if (img_x > 384)
                                        {
                                            img_x = 384;
                                            img_y = avatar_raw.Height - (int)Math.Floor(Math.Floor((double)avatar_raw.Height / 100) * Math.Floor(((double)(avatar_raw.Width - 384) / avatar_raw.Width) * 100));
                                        }

                                        if (img_y > 384)
                                        {
                                            img_x -= (int)Math.Floor(Math.Floor((double)img_x / 100) * Math.Floor(((double)(img_y - 384) / img_y) * 100));
                                            img_y = 384;
                                        }

                                        using (Bitmap avatar_sized = new Bitmap(img_x, img_y))
                                        {
                                            using (Graphics g = Graphics.FromImage(avatar_sized))
                                            {
                                                g.InterpolationMode = InterpolationMode.HighQualityBicubic;
                                                g.DrawImage(avatar_raw, new RectangleF(0, 0, img_x, img_y));

                                                using (MemoryStream ms = new MemoryStream())
                                                {
                                                    avatar_sized.Save(ms, ImageFormat.Jpeg);
                                                    scr.Data = Zip.Compress(ms.ToArray());
                                                }
                                            }
                                        }
                                    }
                                }
                                catch { }
                            }
                        }
                        catch { }
                    }
            }

            return scr;
        }
    }

    public class ScribbleResult : ObjectInstance, ISyncListItem
    {
        public ScribbleResult(ObjectInstance prototype)
            : base(prototype)
        {
            this.PopulateFunctions();
            this.DefineProperty("toAvatar", new PropertyDescriptor(this["toAvatar"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("save", new PropertyDescriptor(this["save"], PropertyAttributes.Sealed), throwOnError: true);
        }

        public byte[] Data { get; set; }
        public UserDefinedFunction Callback { get; set; }
        public String ScriptName { get; set; }
        public String Arg { get; set; }

        [JSProperty(Name = "arg")]
        public String GetArgument
        {
            get { return this.Arg; }
            set { }
        }

        [JSProperty(Name = "exists")]
        public bool DoesExist
        {
            get { return this.Data != null; }
            set { }
        }

        [JSFunction(Name = "toAvatar")]
        public AvatarImage ToAvatar()
        {
            if (this.Data != null)
            {
                AvatarImage a = new AvatarImage(this.Engine.Object.InstancePrototype);
                a.FromScribble(Zip.Decompress(this.Data));
                return a;
            }

            return null;
        }

        private String[] bad_chars = new String[]
        {
            ".",
            "/",
            "\\",
            " ",
        };

        [JSFunction(Name = "save")]
        public bool Save(object a)
        {
            if (this.Data == null)
                return false;

            if (a is String || a is ConcatenatedString)
            {
                String filename = a.ToString();

                if (filename.Length > 1)
                    if (bad_chars.Count<String>(x => filename.Contains(x)) == 0)
                    {
                        String path = Settings.Path + "Scripts\\" + this.Engine.ScriptName + "\\data";

                        try
                        {
                            if (!Directory.Exists(path))
                                Directory.CreateDirectory(path);

                            path += "\\" + filename;

                            File.WriteAllBytes(path, Zip.Decompress(this.Data));
                            return true;
                        }
                        catch { }
                    }
            }

            return false;
        }

        public void FromAvatar(byte[] data)
        {
            this.Data = Zip.Compress(data);
        }

        public void SendScribble(IUser target)
        {
            if (this.Data == null)
                return;

            List<byte> b = new List<byte>(this.Data);

            if (b.Count <= 4000)
            {
                target.SendCustomClientData("cb0t_scribble_once", Settings.BotName, this.Data);
            }
            else
            {
                List<byte[]> p = new List<byte[]>();

                while (b.Count > 4000)
                {
                    p.Add(b.GetRange(0, 4000).ToArray());
                    b.RemoveRange(0, 4000);
                }

                if (b.Count > 0)
                    p.Add(b.ToArray());

                for (int i = 0; i < p.Count; i++)
                {
                    if (i == 0)
                        target.SendCustomClientData("cb0t_scribble_first", "Callisto", p[i]);
                    else if (i == (p.Count - 1))
                        target.SendCustomClientData("cb0t_scribble_last", "Callisto", p[i]);
                    else
                        target.SendCustomClientData("cb0t_scribble_chunk", "Callisto", p[i]);
                }
            }
        }

        public override string ToString()
        {
            return "[object ScribbleResult]";
        }
    }
}
