﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using System.IO;
using System.Data.SQLite;
using Jurassic;
using Jurassic.Library;

namespace callisto.Scripting.Objects
{
    class Sql : ClrFunction
    {
        public Sql(ScriptEngine engine)
            : base(engine.Function.InstancePrototype, "Sql", new SqlInstance(engine.Object.InstancePrototype))
        {

        }

        [JSConstructorFunction]
        public SqlInstance Construct()
        {
            return new SqlInstance(this.InstancePrototype);
        }
    }

    public class SqlInstance : ObjectInstance, IDisposable
    {
        private String m_last_error = String.Empty;

        public SqlInstance(ObjectInstance prototype)
            : base(prototype)
        {
            this.PopulateFunctions();
            this.DefineProperty("open", new PropertyDescriptor(this["open"], Jurassic.Library.PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("close", new PropertyDescriptor(this["close"], Jurassic.Library.PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("query", new PropertyDescriptor(this["query"], Jurassic.Library.PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("value", new PropertyDescriptor(this["value"], Jurassic.Library.PropertyAttributes.Sealed), throwOnError: true);
        }

        private String ConnectionString { get; set; }
        private SQLiteConnection Connection { get; set; }
        private SQLiteDataReader Reader { get; set; }

        private static String[] bad_chars = new String[]
        {
            ".",
            "/",
            "\\",
            " ",
        };

        [JSProperty(Name = "lastError")]
        public String LastError
        {
            get { return this.m_last_error.Replace("SQLite error\r\n", String.Empty); }
            set { }
        }

        [JSProperty(Name = "connected")]
        public bool Connected
        {
            get
            {
                if (this.Connection != null)
                    return this.Connection.State == ConnectionState.Open;

                return false;
            }
            set { }
        }

        [JSProperty(Name = "read")]
        public bool CanRead
        {
            get
            {
                if (this.Reader == null)
                    return false;

                if (!this.Reader.Read())
                    return false;

                return true;
            }
        }

        [JSFunction(Name = "value")]
        public String GetValue(object a)
        {
            this.m_last_error = String.Empty;

            if (a != null)
            {
                if (this.Reader == null)
                {
                    this.m_last_error = "no data available";
                    return null;
                }

                try
                {
                    return this.Reader[a.ToString()].ToString();
                }
                catch (Exception e)
                {
                    this.m_last_error = e.Message;
                    return null;
                }
            }

            this.m_last_error = "invalid item name";
            return null;
        }

        [JSFunction(Name = "query")]
        public bool GetQuery(QueryInstance query)
        {
            this.m_last_error = String.Empty;

            try
            {
                if (this.Reader != null)
                {
                    this.Reader.Close();
                    this.Reader.Dispose();
                    this.Reader = null;
                }
            }
            catch { }

            if (!this.Connected)
            {
                this.m_last_error = "connection closed";
                return false;
            }

            try
            {
                using (SQLiteCommand cmd = new SQLiteCommand(query.query, this.Connection))
                {
                    for (int i = 0; i < query._params.Count; i++)
                        cmd.Parameters.Add(query._params[i]);

                    this.Reader = cmd.ExecuteReader();
                }
            }
            catch (Exception e)
            {
                this.m_last_error = e.Message;
                return false;
            }

            return true;
        }

        [JSFunction(Name = "open")]
        public bool Open(object a)
        {
            if (a == null)
                return false;

            this.LastError = String.Empty;
            String file = a.ToString();

            if (file.Length > 1)
                if (bad_chars.Count<String>(x => file.Contains(x)) == 0)
                {
                    String path = Settings.Path + "Scripts\\" + this.Engine.ScriptName + "\\sql";

                    try
                    {
                        if (!Directory.Exists(path))
                            Directory.CreateDirectory(path);

                        path += "\\" + file;
                        this.ConnectionString = "Data Source=\"" + path + "\"";

                        if (!File.Exists(path))
                            SQLiteConnection.CreateFile(path);
                    }
                    catch { return false; }

                    try
                    {
                        if (this.Reader != null)
                        {
                            this.Reader.Close();
                            this.Reader.Dispose();
                            this.Reader = null;
                        }
                    }
                    catch { }

                    try
                    {
                        if (this.Connection == null)
                            this.Connection = new SQLiteConnection(this.ConnectionString);

                        if (this.Connection.State != ConnectionState.Open)
                            this.Connection.Open();

                        return this.Connection.State == ConnectionState.Open;
                    }
                    catch { }
                }

            return false;
        }

        [JSFunction(Name = "close")]
        public bool Close()
        {
            this.m_last_error = String.Empty;
            bool success = true;

            try
            {
                if (this.Connection != null)
                {
                    this.Connection.Close();
                    this.Connection.Dispose();
                    this.Connection = null;
                }
            }
            catch { success = false; }

            try
            {
                if (this.Reader != null)
                {
                    this.Reader.Close();
                    this.Reader.Dispose();
                    this.Reader = null;
                }
            }
            catch { }

            return success;
        }

        public void Dispose()
        {
            try
            {
                if (this.Connection != null)
                {
                    this.Connection.Close();
                    this.Connection.Dispose();
                    this.Connection = null;
                }
            }
            catch { }

            try
            {
                if (this.Reader != null)
                {
                    this.Reader.Close();
                    this.Reader.Dispose();
                    this.Reader = null;
                }
            }
            catch { }
        }
    }
}
