﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using System.IO;
using Jurassic;
using Jurassic.Library;

namespace callisto.Scripting.Objects
{
    class XmlParser : ClrFunction
    {
        public XmlParser(ScriptEngine engine)
            : base(engine.Function.InstancePrototype, "XmlParser", new XmlParserInstance(engine.Object.InstancePrototype))
        {

        }

        [JSConstructorFunction]
        public XmlParserInstance Construct()
        {
            return new XmlParserInstance(this.InstancePrototype);
        }
    }

    class XmlParserInstance : ObjectInstance
    {
        private XmlDocument DOC { get; set; }
        private bool _avail = false;

        public XmlParserInstance(ObjectInstance prototype)
            : base(prototype)
        {
            this.DOC = new XmlDocument();
            this.PopulateFunctions();
            this.DefineProperty("load", new PropertyDescriptor(this["load"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("create", new PropertyDescriptor(this["create"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("getNodesByName", new PropertyDescriptor(this["getNodesByName"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("appendChild", new PropertyDescriptor(this["appendChild"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("removeChild", new PropertyDescriptor(this["removeChild"], PropertyAttributes.Sealed), throwOnError: true);
        }

        [JSProperty(Name = "available")]
        public bool Available
        {
            get { return this._avail; }
            set { }
        }

        [JSProperty(Name = "xml")]
        public String Xml
        {
            get
            {
                if (!this._avail)
                    return null;

                try
                {
                    XmlDocument doc = new XmlDocument();
                    doc.LoadXml(this.DOC.InnerXml);
                    String result = String.Empty;

                    using (MemoryStream ms = new MemoryStream())
                    {
                        XmlTextWriter writer = new XmlTextWriter(ms, Encoding.UTF8);
                        writer.Formatting = Formatting.Indented;
                        doc.Save(writer);
                        ms.Position = 0;

                        using (StreamReader reader = new StreamReader(ms, true))
                            result = reader.ReadToEnd();
                    }

                    return result;
                }
                catch { return null; }
            }
            set { }
        }

        [JSProperty(Name = "nodeName")]
        public String Name
        {
            get
            {
                if (!this._avail)
                    return null;

                try
                {
                    return this.DOC.Name;
                }
                catch { return null; }
            }
            set { }
        }

        [JSProperty(Name = "nodeValue")]
        public String Value
        {
            get
            {
                if (!this._avail)
                    return null;

                try
                {
                    return this.DOC.InnerText;
                }
                catch { return null; }
            }
            set { this.DOC.InnerText = value; }
        }

        [JSProperty(Name = "parentNode")]
        public Node ParentNode
        {
            get
            {
                if (!this._avail)
                    return null;

                try
                {
                    if (this.DOC.ParentNode != null)
                        return new Node(this.Engine.Object.InstancePrototype, this.DOC.ParentNode);
                }
                catch { }

                return null;
            }
            set { }
        }

        [JSProperty(Name = "childNodes")]
        public NodeCollection ChildNodes
        {
            get
            {
                if (!this._avail)
                    return null;

                try
                {
                    if (this.DOC.ChildNodes.Count > 0)
                        return new NodeCollection(this.Engine.Object.InstancePrototype, this.DOC.ChildNodes);
                }
                catch { }

                return null;
            }
            set { }
        }

        [JSProperty(Name = "attributes")]
        public NodeAttributes Attributes
        {
            get
            {
                if (!this._avail)
                    return null;

                try
                {
                    return new NodeAttributes(this.Engine.Object.InstancePrototype, this.DOC.Attributes, this.DOC);
                }
                catch { }

                return null;
            }
            set { }
        }

        [JSFunction(Name = "load")]
        public bool Load(object a)
        {
            if (a != null)
            {
                try
                {
                    this.DOC = new XmlDocument();
                    this.DOC.PreserveWhitespace = true;
                    this.DOC.LoadXml(a.ToString());
                    this._avail = true;
                    return true;
                }
                catch { this._avail = false; }
            }

            return false;
        }

        [JSFunction(Name = "create")]
        public bool Create(object a)
        {
            if (a == null)
                return false;

            try
            {
                this.DOC = new XmlDocument();
                this.DOC.PreserveWhitespace = true;
                this.DOC.AppendChild(this.DOC.CreateXmlDeclaration("1.0", null, null));
                this.DOC.AppendChild(this.DOC.CreateElement(a.ToString()));
                this._avail = true;
                return true;
            }
            catch { }

            this._avail = false;
            return false;
        }

        [JSFunction(Name = "getNodesByName")]
        public NodeCollection GetNodesByName(object a)
        {
            if (a != null)
                try
                {
                    XmlNodeList list = this.DOC.GetElementsByTagName(a.ToString());

                    if (list.Count > 0)
                        return new NodeCollection(this.Engine.Object.InstancePrototype, list);
                }
                catch { }

            return null;
        }

        [JSFunction(Name = "appendChild")]
        public Node AppendChild(object a)
        {
            if (a == null)
                return null;

            try
            {
                XmlNode x = this.DOC.CreateNode(XmlNodeType.Element, a.ToString(), this.DOC.DocumentElement.BaseURI);
                this.DOC.DocumentElement.AppendChild(x);
                return new Node(this.Engine.Object.InstancePrototype, x);
            }
            catch { }

            return null;
        }

        [JSFunction(Name = "removeChild")]
        public bool RemoveChild(object a)
        {
            if (a is Node)
            {
                try
                {
                    Node n = (Node)a;
                    this.DOC.DocumentElement.RemoveChild(n.Item);
                    return true;
                }
                catch { }
            }

            return false;
        }
    }

    class NodeCollection : ObjectInstance
    {
        private XmlNodeList Items { get; set; }
        private int count { get; set; }

        public NodeCollection(ObjectInstance prototype, XmlNodeList list)
            : base(prototype)
        {
            this.Items = list;
            this.count = 0;

            this.PopulateFunctions();

            foreach (XmlNode n in list)
                if (!n.Name.StartsWith("#"))
                    this.SetPropertyValue((uint)this.count++,
                        new Node(this.Engine.Object.InstancePrototype, n), throwOnError: true);
        }

        [JSProperty(Name = "length")]
        public int Length
        {
            get { return this.count; }
            set { }
        }

        public override string ToString()
        {
            return "[object NodeCollection]";
        }
    }

    class Node : ObjectInstance
    {
        internal XmlNode Item { get; set; }

        public Node(ObjectInstance prototype, XmlNode node)
            : base(prototype)
        {
            this.Item = node;
            this.PopulateFunctions();
            this.DefineProperty("getNodesByName", new PropertyDescriptor(this["getNodesByName"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("appendChild", new PropertyDescriptor(this["appendChild"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("removeChild", new PropertyDescriptor(this["removeChild"], PropertyAttributes.Sealed), throwOnError: true);
        }

        [JSProperty(Name = "nodeName")]
        public String Name
        {
            get
            {
                try
                {
                    return this.Item.Name;
                }
                catch { return null; }
            }
            set { }
        }

        [JSProperty(Name = "nodeValue")]
        public String Value
        {
            get
            {
                try
                {
                    return this.Item.InnerText;
                }
                catch { return null; }
            }
            set { this.Item.InnerText = value; }
        }

        [JSProperty(Name = "parentNode")]
        public Node ParentNode
        {
            get
            {
                try
                {
                    if (this.Item.ParentNode != null)
                        return new Node(this.Engine.Object.InstancePrototype, this.Item.ParentNode);
                }
                catch { }

                return null;
            }
            set { }
        }

        [JSProperty(Name = "childNodes")]
        public NodeCollection ChildNodes
        {
            get
            {
                try
                {
                    if (this.Item.ChildNodes.Count > 0)
                        return new NodeCollection(this.Engine.Object.InstancePrototype, this.Item.ChildNodes);
                }
                catch { }

                return null;
            }
            set { }
        }

        [JSProperty(Name = "attributes")]
        public NodeAttributes Attributes
        {
            get
            {
                try
                {
                    return new NodeAttributes(this.Engine.Object.InstancePrototype, this.Item.Attributes, this.Item);
                }
                catch { }

                return null;
            }
            set { }
        }

        [JSFunction(Name = "getNodesByName")]
        public NodeCollection GetNodesByName(object a)
        {
            if (a != null)
                try
                {
                    XmlElement e = (XmlElement)this.Item;
                    XmlNodeList list = e.GetElementsByTagName(a.ToString());
                    
                    if (list.Count > 0)
                        return new NodeCollection(this.Engine.Object.InstancePrototype, list);
                }
                catch { }

            return null;
        }

        [JSFunction(Name = "appendChild")]
        public Node AppendChild(object a)
        {
            if (a == null)
                return null;

            try
            {
                XmlNode x = this.Item.OwnerDocument.CreateNode(XmlNodeType.Element, a.ToString(), this.Item.BaseURI);
                this.Item.AppendChild(x);
                return new Node(this.Engine.Object.InstancePrototype, x);
            }
            catch { }

            return null;
        }

        [JSFunction(Name = "removeChild")]
        public bool RemoveChild(object a)
        {
            if (a is Node)
            {
                try
                {
                    Node n = (Node)a;
                    this.Item.RemoveChild(n.Item);
                    return true;
                }
                catch { }
            }

            return false;
        }

        public override string ToString()
        {
            return "[object Node]";
        }
    }

    class NodeAttributes : ObjectInstance
    {
        private XmlAttributeCollection Attribs { get; set; }
        private XmlNode Owner { get; set; }
        private int count { get; set; }

        public NodeAttributes(ObjectInstance prototype, XmlAttributeCollection attribs, XmlNode owner)
            : base(prototype)
        {
            this.Attribs = attribs;
            this.Owner = owner;
            this.count = 0;

            this.PopulateFunctions();

            this.DefineProperty("getValue", new PropertyDescriptor(this["getValue"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("setValue", new PropertyDescriptor(this["setValue"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("removeValue", new PropertyDescriptor(this["removeValue"], PropertyAttributes.Sealed), throwOnError: true);

            foreach (XmlAttribute a in attribs)
                this.SetPropertyValue((uint)this.count++, a.Name, throwOnError: true);
        }

        [JSProperty(Name = "length")]
        public int Length
        {
            get { return this.count; }
            set { }
        }

        [JSFunction(Name = "getValue")]
        public String GetValue(object a)
        {
            if (a == null)
                return null;

            try
            {
                XmlAttribute att = this.Attribs[a.ToString()];

                if (att != null)
                    return att.InnerText;
            }
            catch { }

            return null;
        }

        [JSFunction(Name = "setValue")]
        public bool SetValue(object a, object b)
        {
            if (a == null || b == null)
                return false;

            try
            {
                String name = a.ToString();
                String text = b.ToString();

                if (this.Attribs[name] != null)
                    this.Attribs[name].InnerText = text;
                else
                {
                    XmlAttribute xa = this.Owner.OwnerDocument.CreateAttribute(name);
                    xa.InnerText = text;
                    this.Owner.Attributes.Append(xa);
                }

                for (int i = 0; i < this.count; i++)
                    this.SetPropertyValue((uint)i, null, throwOnError: true);

                this.count = 0;

                foreach (XmlAttribute att in this.Attribs)
                    this.SetPropertyValue((uint)this.count++, att.Name, throwOnError: true);

                return true;
            }
            catch { }

            return false;
        }

        [JSFunction(Name = "removeValue")]
        public bool RemoveValue(object a)
        {
            if (a == null)
                return false;

            try
            {
                if (this.Attribs[a.ToString()] != null)
                {
                    this.Attribs.Remove(this.Attribs[a.ToString()]);

                    for (int i = 0; i < this.count; i++)
                        this.SetPropertyValue((uint)i, null, throwOnError: true);

                    this.count = 0;

                    foreach (XmlAttribute att in this.Attribs)
                        this.SetPropertyValue((uint)this.count++, att.Name, throwOnError: true);

                    return true;
                }
            }
            catch { }

            return false;
        }

        public override string ToString()
        {
            return "[object NodeAttributes]";
        }
    }
}
