﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using Jurassic;
using Jurassic.Library;

namespace callisto.Scripting.Statics
{
    class Script : ObjectInstance
    {
        public Script(ScriptEngine engine)
            : base(engine)
        {
            this.PopulateFunctions();
            this.DefineProperty("name", new PropertyDescriptor(this["name"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("create", new PropertyDescriptor(this["create"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("load", new PropertyDescriptor(this["load"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("write", new PropertyDescriptor(this["write"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("writeInclude", new PropertyDescriptor(this["writeInclude"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("eval", new PropertyDescriptor(this["eval"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("include", new PropertyDescriptor(this["include"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("kill", new PropertyDescriptor(this["kill"], PropertyAttributes.Sealed), throwOnError: true);
            this.DefineProperty("includeScripts", new PropertyDescriptor(this["includeScripts"], PropertyAttributes.Sealed), throwOnError: true);
        }

        private static String[] bad_chars = new String[]
        {
            "..",
            "/",
            "\\",
            " ",
        };

        [JSFunction(Name = "name", Flags = JSFunctionFlags.HasEngineParameter)]
        public static String Name(ScriptEngine eng)
        {
            return eng.ScriptName;
        }

        [JSFunction(Name = "create")]
        public static bool Create(object name)
        {
            if (name is Undefined || name == null)
                return false;

            String str = name.ToString();

            if (str.Length > 0)
            {
                ScriptManager.Unload(str);
                CallistoScript script = new CallistoScript(str);

                foreach (UserObject u in UserPool.Pool)
                    if (u.Connected)
                        if (u.logged_in)
                            script.AddUser(u);

                foreach (WebUser u in WebPool.Pool)
                    if (u.InUse)
                        if (u.LoggedIn)
                            script.AddUser(u);

                if (CallistoServer.Linker.IsLinking)
                    foreach (LinkedUserObject u in CallistoServer.Linker.linked_users)
                        script.AddUser(u);

                ScriptManager.Scripts.Add(script);
                return true;
            }

            return false;
        }

        [JSFunction(Name = "load")]
        public static bool Load(object name)
        {
            if (name is Undefined || name == null)
                return false;

            String filename = name.ToString();

            if (filename.Length > 3 && filename.EndsWith(".js"))
                if (bad_chars.Count<String>(x => filename.Contains(x)) == 0)
                    return ScriptManager.Load(new FileInfo(Settings.Path + "Scripts\\" + filename + "\\" + filename));

            return false;
        }

        [JSFunction(Name = "write")]
        public static bool Save(object name, object src)
        {
            if (name is Undefined || src is Undefined || name == null || src == null)
                return false;

            String filename = name.ToString();
            String source = src.ToString();

            if (filename.Length > 3 && filename.EndsWith(".js"))
                if (bad_chars.Count<String>(x => filename.Contains(x)) == 0)
                {
                    try
                    {
                        if (!Directory.Exists(Settings.Path + "Scripts\\" + filename))
                            Directory.CreateDirectory(Settings.Path + "Scripts\\" + filename);

                        System.IO.File.WriteAllText(Settings.Path + "Scripts\\" + filename + "\\" + filename, source);
                        return true;
                    }
                    catch { }
                }

            return false;
        }

        [JSFunction(Name = "writeInclude")]
        public static bool SaveInclude(object name, object inc, object src)
        {
            if (name is Undefined || inc is Undefined || src is Undefined || name == null || inc == null || src == null)
                return false;

            String filename = name.ToString();
            String include = inc.ToString();
            String source = src.ToString();

            if (filename.Length > 3 && filename.EndsWith(".js"))
                if (bad_chars.Count<String>(x => filename.Contains(x)) == 0)
                    if (include.Length > 3 && include.EndsWith(".js"))
                        if (bad_chars.Count<String>(x => include.Contains(x)) == 0)
                        {
                            try
                            {
                                if (!Directory.Exists(Settings.Path + "Scripts\\" + filename))
                                    Directory.CreateDirectory(Settings.Path + "Scripts\\" + filename);

                                System.IO.File.WriteAllText(Settings.Path + "Scripts\\" + filename + "\\" + include, source);
                                return true;
                            }
                            catch { }
                        }

            return false;
        }

        [JSFunction(Name = "eval")]
        public static String Eval(object name, object src)
        {
            if (name is Undefined || src is Undefined || name == null || src == null)
                return null;

            String str = name.ToString();
            CallistoScript script = ScriptManager.Scripts.Find(x => x.ScriptName == str);

            if (script != null)
            {
                str = src.ToString();

                try
                {
                    object result = script.JS.Evaluate(str);

                    if (result is bool)
                        return result.ToString().ToLower();
                    else
                        return result.ToString();
                }
                catch (JavaScriptException e)
                {
                    ScriptManager.OnError(script.ScriptName, e.Message, e.LineNumber);
                }
                catch { }
            }

            return null;
        }

        [JSFunction(Name = "include", Flags = JSFunctionFlags.HasEngineParameter)]
        public static bool Include(ScriptEngine eng, object inc, object fdlr)
        {
            if (inc is Undefined || inc == null || fdlr is Undefined || fdlr == null)
                return false;

            String folder = fdlr.ToString();
            String include = inc.ToString();
            String source = null;

            if (include.Length > 3 && include.EndsWith(".js") && folder.Length > 3)
                if (bad_chars.Count<String>(x => include.Contains(x)) == 0 && bad_chars.Count<String>(x => folder.Contains(x)) == 0)
                {
                    try
                    {
                        source = System.IO.File.ReadAllText(Settings.Path + "Scripts\\" + eng.ScriptName + "\\" + folder + "\\" + include);
                    }
                    catch { }
                }

            if (source != null)
            {
                CallistoScript script = ScriptManager.Scripts.Find(x => x.ScriptName == eng.ScriptName);

                if (script != null)
                {
                    try
                    {
                        script.JS.Evaluate(source);
                        return true;
                    }
                    catch { }
                }
            }

            return false;
        }

        [JSFunction(Name = "include", Flags = JSFunctionFlags.HasEngineParameter)]
        public static bool Include(ScriptEngine eng, object inc)
        {
            if (inc is Undefined || inc == null)
                return false;

            String include = inc.ToString();
            String source = null;

            if (include.Length > 3 && include.EndsWith(".js"))
                if (bad_chars.Count<String>(x => include.Contains(x)) == 0)
                {
                    try
                    {
                        source = System.IO.File.ReadAllText(Settings.Path + "Scripts\\" + eng.ScriptName + "\\" + include);
                    }
                    catch { }
                }

            if (source != null)
            {
                CallistoScript script = ScriptManager.Scripts.Find(x => x.ScriptName == eng.ScriptName);

                if (script != null)
                {
                    try
                    {
                        script.JS.Evaluate(source);
                        return true;
                    }
                    catch { }
                }
            }

            return false;
        }

        [JSFunction(Name = "includeScripts", Flags = JSFunctionFlags.HasEngineParameter)]
        public static bool IncludeScripts(ScriptEngine eng, object inc)
        {
            if (inc is Undefined || inc == null)
                return false;

            String include = inc.ToString();
            String[] scripts = null;

            if (include.Length > 3)
            {
                if (bad_chars.Count<String>(x => include.Contains(x)) == 0)
                {
                    try
                    {
                        scripts = Directory.GetFiles(Settings.Path + "Scripts\\" + eng.ScriptName + "\\" + include + "\\", "*.js");
                    }
                    catch { }
                }
            }
            if(scripts != null) 
            {
                CallistoScript script = ScriptManager.Scripts.Find(x => x.ScriptName == eng.ScriptName);
                if(script != null) 
                {
                    foreach (String s in scripts)
                    {
                        try
                        {
                            String source = System.IO.File.ReadAllText(s);
                            if (source != null)
                            {
                                script.JS.Evaluate(source);
                            }
                        }
                        catch { }
                    }
                    return true;
                }
            }
            return false;
        }

        [JSFunction(Name = "kill")]
        public static bool Kill(object name)
        {
            if (name is Undefined || name == null)
                return false;

            return ScriptManager.Unload(name.ToString());
        }

    }
}
