﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Security.Cryptography;
using ICallisto;

namespace callisto
{
    class ChatPacketProcessor
    {
        public static void Eval(UserObject userobj, ProtoMessage msg, AresPacketReader packet, uint time)
        {
            if (!userobj.logged_in)
            {
                if (msg != ProtoMessage.MSG_SERVER_TOHUB_LOGINREQ)
                {
                    if (msg != ProtoMessage.MSG_CHAT_CLIENT_LOGIN && msg != ProtoMessage.MSG_CHAT_CLIENT_RELOGIN)
                    {
                        CallistoPluginStats.rejection_count++;
                        throw new Exception("user sending other packets before logging in first!");
                    }
                }
            }

            if (userobj.IsFlooding(msg, packet.ToArray(), time))
            {
                if (ServerEvents.OnFlood(userobj, msg))
                {
                    CallistoPluginStats.flood_count++;
                    throw new Exception("user disconnected for flooding");
                }
            }

            switch (msg)
            {
                case ProtoMessage.MSG_CHAT_CLIENT_RELOGIN:
                    ProcessLogin(userobj, packet, true, time);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_LOGIN:
                    ProcessLogin(userobj, packet, false, time);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_UPDATE_STATUS:
                    userobj.last_update_time = time;
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_PVT:
                    ProcessPrivateMessage(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_PUBLIC:
                    ProcessPublicMessage(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_EMOTE:
                    ProcessEmoteMessage(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_FASTPING:
                    userobj.fastping = true;
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_COMMAND:
                    ProcessCommandMessage(userobj, packet.ReadString());
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_AUTHLOGIN:
                    ServerEvents.OnCommand(userobj, "login " + packet.ReadString(), null, String.Empty);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_AVATAR:
                    ProcessAvatar(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_PERSONAL_MESSAGE:
                    String text = Helpers.FormatAresColorCodes(packet.ReadString());

                    if (ServerEvents.OnPersonalReceived(userobj, text))
                        userobj.PersonalMessage = text;

                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_ADDSHARE:
                    ProcessAddSharedItem(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_REMSHARE:
                    ProcessRemoveSharedItem(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_SEARCH:
                    ProcessSearchRequest(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_BROWSE:
                    ProcessBrowseRequest(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_DUMMY:
                    ProcessDummy(userobj);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_DIRCHATPUSH:
                    ProcessDirectChatPush(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_SEND_SUPERNODES:
                    userobj.SendPacket(ChatPacketHelpers.SuperNodes(UserPool.GetSuperNodes()));
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_IGNORELIST:
                    ProcessIgnoreRequest(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_AUTOLOGIN:
                    ProcessSHA1Login(userobj, packet);
                    break;

                case ProtoMessage.MSG_SERVER_TOHUB_LOGINREQ:
                    ProcessLinkRequest(userobj, packet, time);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_CUSTOM_DATA:
                    ProcessCustomData(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_CUSTOM_DATA_ALL:
                    ProcessCustomDataAll(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_ADVANCED_FEATURES_PROTOCOL:
                    ushort adf_len = packet.ReadInt16();
                    ProtoMessage adf_id = (ProtoMessage)packet.ReadByte();
                    byte[] adf_data = packet.ReadBytes();

                    if (adf_data.Length == adf_len)
                        EvalAdvancedFeatures(userobj, adf_id, new AresPacketReader(adf_data));

                    break;
            }
        }

        public static void EvalAdvancedFeatures(UserObject userobj, ProtoMessage msg, AresPacketReader packet)
        {
            switch (msg)
            {
                case ProtoMessage.MSG_CHAT_CLIENT_CUSTOM_ADD_TAGS:
                    ProcessCustomDataAddTags(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_CUSTOM_REM_TAGS:
                    ProcessCustomDataRemTags(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_CUSTOM_FONT:
                    ProcessCustomFont(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_VC_SUPPORTED:
                    userobj.voice_chat = packet.ReadByte() == 1;
                    userobj.voice_private_chat = packet.ReadByte() == 1;
                    UserPool.BroadcastToVroom(userobj.vroom, ChatPacketHelpers.VoiceChatUserSupport(userobj));
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_VC_FIRST:
                    ProcessVCFirst(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_VC_FIRST_TO:
                    ProcessVCFirstTo(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_VC_CHUNK:
                    ProcessVCChunk(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_VC_CHUNK_TO:
                    ProcessVCChunkTo(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_VC_IGNORE:
                    ProcessVCIgnore(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_SUPPORTS_CUSTOM_EMOTES:
                    ProcessSupportsCustomEmotes(userobj);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_CUSTOM_EMOTES_UPLOAD_ITEM:
                    ProcessCustomEmoteUpload(userobj, packet);
                    break;

                case ProtoMessage.MSG_CHAT_CLIENT_CUSTOM_EMOTE_DELETE:
                    ProcessCustomEmoteDelete(userobj, packet);
                    break;
            }
        }

        private static void ProcessSHA1Login(UserObject userobj, AresPacketReader packet)
        {
            if (PasswordManager.IsSHA1Login(userobj, packet.ReadBytes()))
                ServerEvents.OnLoginRequested(userobj);
        }

        private static void ProcessLinkRequest(UserObject userobj, AresPacketReader packet, uint time)
        {
            if (CallistoServer.Linker.IsActive || !Settings.RoomLinking) // go away
            {
                userobj.SendPacket(LinkPacketHelpers.LinkError(LinkError.LinkingDisabled));
                userobj.Disconnect();
                return;
            }

            CallistoServer.Linker = new LinkClient(userobj.socket, time);
            userobj.socket = null;
            packet.SkipBytes(2); // user count - don't care
            CallistoServer.Linker.protocol = packet.ReadInt16();
            CallistoServer.Linker.target_port = packet.ReadInt16();
            CallistoServer.Linker.room_name = packet.ReadString();
            CallistoServer.Linker.BuildHashlink();
            userobj = new UserObject();

            if (CallistoServer.Linker.protocol < Settings.LINK_PROTO)
            {
                CallistoServer.Linker.SendPacket(LinkPacketHelpers.LinkError(LinkError.OldProtocol));
                CallistoServer.Linker.SendAndDisconnect();
                return;
            }

            if (Helpers.IsLocalHost(CallistoServer.Linker.target_ip) && ((ushort)Settings.ServerPort) == CallistoServer.Linker.target_port)
            {
                CallistoServer.Linker.SendPacket(LinkPacketHelpers.LinkError(LinkError.LinkToSelf));
                CallistoServer.Linker.SendAndDisconnect();
                return;
            }

            var istrusted = from x in TrustedServers.Items
                       where x.Name == CallistoServer.Linker.Name
                           && ((ushort)x.Port) == CallistoServer.Linker.Port
                           && x.IP.GetAddressBytes()[0] == CallistoServer.Linker.ExternalIP.GetAddressBytes()[0]
                           && x.IP.GetAddressBytes()[1] == CallistoServer.Linker.ExternalIP.GetAddressBytes()[1]
                       select x;

            if (istrusted.Count<TrustedServerItems>() == 0)
            {
                CallistoServer.Linker.SendPacket(LinkPacketHelpers.LinkError(LinkError.NotOnAllowList));
                CallistoServer.Linker.SendAndDisconnect();
                return;
            }

            CallistoServer.Linker.SendPacket(LinkPacketHelpers.RequestLinkAckPacket());
            UserPool.UncloakAll();
            UserPool.SendUserlistToLink(CallistoServer.Linker);
            CallistoServer.Linker.logged_in = true;
            UserPool.SendAvatarsToLink(CallistoServer.Linker);
            UserPool.SendCustomTagsToLink(CallistoServer.Linker);
            UserPool.SendCustomNamesToLink(CallistoServer.Linker);
            ServerEvents.OnLinkJoin(CallistoServer.Linker);
        }

        private static void ProcessDummy(UserObject userobj)
        {
            ServerEvents.OnRejected(userobj, NativeRejectionType.Banned);
            Proxies.AddNonTorProxy(userobj);
            userobj.Ban();
        }

        private static void ProcessLogin(UserObject userobj, AresPacketReader packet, bool relogin, uint tt)
        {
            if (userobj.logged_in)
                throw new Exception("user disconnected for login flooding");

            userobj.PopulateCredentials(packet);

            if (Encoding.UTF8.GetByteCount(userobj.version) > 16)
                throw new Exception("invalid credentials");

            UserObject ghost_user1 = null;
            WebUser ghost_user2 = null;

            if (Proxies.IsProxy(userobj))
            {
                ServerEvents.OnRejected(userobj, NativeRejectionType.Banned);
                userobj.Ban();
                return;
            }

            if (UserRecordManager.IsMaxClientsExcess(userobj, 4))
            {
                ServerEvents.OnRejected(userobj, NativeRejectionType.ClientFlood);
                userobj.Disconnect();
                return;
            }

            if (UserRecordManager.IsJoinFlooding(userobj) && !userobj.local_host)
            {
                ServerEvents.OnRejected(userobj, NativeRejectionType.LoginFlood);
                userobj.Disconnect();
                return;
            }

            if (UserPool.NameExists(userobj)) // let's prevent name hijacking
            {
                ghost_user1 = UserPool.FindGhost(userobj);

                if (ghost_user1 != null)
                {
                    ghost_user1.Disconnect(true);
                }
                else
                {
                    ghost_user2 = WebPool.FindGhost(userobj);

                    if (ghost_user2 != null)
                        ghost_user2.Disconnect(true);
                    else
                    {
                        ServerEvents.OnRejected(userobj, NativeRejectionType.NameInUse);
                        userobj.Disconnect();
                        return;
                    }
                }
            }

            UserRecordManager.AddRecord(userobj);

            if (BanManager.IsBanned(userobj))
            {
                ServerEvents.OnRejected(userobj, NativeRejectionType.Banned);

                if (ghost_user1 != null)
                    ghost_user1.ShowUserParting();

                if (ghost_user2 != null)
                    ghost_user2.ShowUserParting();

                userobj.Disconnect();
                return;
            }

            if (!ServerEvents.OnJoinCheck(userobj) && !userobj.local_host)
            {
                ServerEvents.OnRejected(userobj, NativeRejectionType.Other);
                CallistoPluginStats.rejection_count++;

                if (ghost_user1 != null)
                    ghost_user1.ShowUserParting();

                if (ghost_user2 != null)
                    ghost_user2.ShowUserParting();

                userobj.Disconnect();
                return;
            }
            else
            {
                userobj.Login(ghost_user1 != null || ghost_user2 != null, relogin);

                if (ghost_user1 == null && ghost_user2 == null)
                    if (CallistoServer.Linker.IsLinking && userobj.vroom == 0)
                        CallistoServer.Linker.SendPacket(ChatPacketHelpers.Join(userobj, true));

                ServerEvents.OnJoin(userobj);

                if (Settings.ban_spambots)
                {
                    if (userobj.Level == AdminLevel.Regular && !Captcha.IsTrusted(userobj.guid))
                    {
                        ServerEvents.OnCaptcha(userobj, CaptchaState.Sending);
                        CaptchaObj cap = Captcha.Create(userobj);

                        foreach (byte[] b in cap.packets)
                            userobj.SendPacket(b);

                        userobj.captcha_answer = cap.answer;
                    }
                    else userobj.has_captched = true;
                }
                else userobj.has_captched = true;
            }
        }

        private static void ProcessPrivateMessage(UserObject userobj, AresPacketReader packet)
        {
            if (!userobj.has_captched)
                return;

            CallistoPluginStats.private_messages++;
            String name = packet.ReadString();
            String text = packet.ReadString().Replace("\x0002", String.Empty);
            UserObject target = UserPool.FindUserByName(name);

            if (name == Settings.BotName)
            {
                if (!(text.StartsWith("#login") || text.StartsWith("/login")))
                    ServerEvents.OnBotPM(userobj, text);

                if (text.StartsWith("#") || text.StartsWith("/"))
                    ProcessCommandMessage(userobj, text.Substring(1));
            }
            else
            {
                if (target == null) // can't find user, maybe a linked user?
                {
                    if (CallistoServer.Linker.IsLinking && userobj.vroom == 0)
                    {
                        if (CallistoServer.Linker.linked_users.Find(l => l.name == name) != null)
                            CallistoServer.Linker.SendPacket(LinkPacketHelpers.RelayToUser(name, ChatPacketHelpers.Private(userobj.name, text)));
                    }
                    else { userobj.SendPacket(ChatPacketHelpers.OfflineUser(name)); } // no such user
                }
                else
                {
                    if (target.ignores.Contains(userobj.name)) // you're being ignored
                    {
                        userobj.SendPacket(ChatPacketHelpers.IsIgnoringYou(target.name));
                    }
                    else // send pm to the target
                    {
                        if (target.cloaked && target.level > userobj.level)
                            userobj.SendPacket(ChatPacketHelpers.OfflineUser(name));
                        else
                        {
                            text = ServerEvents.OnPM(userobj, target, text);

                            if (text.Length > 0 && userobj.logged_in) // blocked pm will be empty by now
                                target.SendPacket(ChatPacketHelpers.Private(userobj.name, text));
                        }
                    }
                }
            }
        }

        private static void ProcessPublicMessage(UserObject userobj, AresPacketReader packet)
        {
            if (!userobj.has_captched)
            {
                String cap_str = packet.ReadString();
                cap_str = Helpers.StripColors(cap_str);
                cap_str = cap_str.Trim();

                if (cap_str.Length > 0)
                {
                    userobj.SendPacket(ChatPacketHelpers.Public(userobj.name, cap_str));
                    int cap;

                    if (int.TryParse(cap_str, out cap))
                    {
                        if (userobj.captcha_answer == cap)
                        {
                            userobj.has_captched = true;
                            ServerEvents.OnCaptcha(userobj, CaptchaState.Success);
                            Captcha.MakeTrusted(userobj.guid);
                            return;
                        }
                    }
                }

                userobj.captcha_attempts++;

                if (userobj.captcha_attempts > 2)
                {
                    ServerEvents.OnCaptcha(userobj, CaptchaState.Exceeded);
                    throw new Exception("captcha");
                }
                else
                {
                    ServerEvents.OnCaptcha(userobj, CaptchaState.Incorrect);
                    CaptchaObj cap = Captcha.Create(userobj);

                    foreach (byte[] b in cap.packets)
                        userobj.SendPacket(b);

                    userobj.captcha_answer = cap.answer;
                    return;
                }
            }

            CallistoPluginStats.public_messages++;
            String text = packet.ReadString().Replace("\x0002", String.Empty);

            if (text.StartsWith("#"))
                ProcessCommandMessage(userobj, text.Substring(1));

            ServerEvents.OnTextReceived(userobj, text);

            if (userobj.logged_in)
            {
                if (CallistoServer.Linker.IsLinking && userobj.vroom == 0)
                    CallistoServer.Linker.SendPacket(LinkPacketHelpers.TextReceived(userobj, text));

                text = ServerEvents.OnTextBefore(userobj, text);

                if (text.Length > 0 && userobj.logged_in && !userobj.Muzzled) // blocked text will be empty by now
                {
                    if (userobj.CustomName.Length > 0 && Settings.CAN_CUSTOM_NAME)
                    {
                        UserPool.BroadcastToUnignored(userobj, ChatPacketHelpers.NoSuch(userobj.CustomName + text));
                        WebPool.BroadcastCustomNamePublic(userobj, text);
                    }
                    else
                    {
                        UserPool.BroadcastToUnignored(userobj, ChatPacketHelpers.Public(userobj.name, text));
                        WebPool.BroadcastPublic(userobj, text);
                    }

                    ServerEvents.OnTextAfter(userobj, text);

                    if (CallistoServer.Linker.IsLinking && userobj.vroom == 0)
                        CallistoServer.Linker.SendPacket(ChatPacketHelpers.Public(userobj.name, text));
                }
            }
        }

        private static void ProcessEmoteMessage(UserObject userobj, AresPacketReader packet)
        {
            if (!userobj.has_captched)
                return;

            CallistoPluginStats.public_messages++;
            String text = packet.ReadString().Replace("\x0002", String.Empty);
            ServerEvents.OnEmoteReceived(userobj, text);

            if (userobj.logged_in)
            {
                if (CallistoServer.Linker.IsLinking && userobj.vroom == 0)
                    CallistoServer.Linker.SendPacket(LinkPacketHelpers.EmoteReceived(userobj, text));

                text = ServerEvents.OnEmoteBefore(userobj, text);

                if (text.Length > 0 && userobj.logged_in && !userobj.Muzzled) // blocked text will be empty by now
                {
                    UserPool.BroadcastToUnignored(userobj, ChatPacketHelpers.Emote(userobj.name, text));
                    WebPool.BroadcastEmote(userobj, text);
                    ServerEvents.OnEmoteAfter(userobj, text);

                    if (CallistoServer.Linker.IsLinking && userobj.vroom == 0)
                        CallistoServer.Linker.SendPacket(ChatPacketHelpers.Emote(userobj.name, text));
                }
            }
        }

        private static void ProcessCommandMessage(UserObject userobj, String text)
        {
            if (!userobj.has_captched)
                return;

            CommandObject cobj = Helpers.EvaluateClientCommandString(text);

            if (CallistoServer.Linker.IsLinking && cobj.tUser != null)
            {
                if (cobj.tUser.IsLinked)
                {
                    CallistoServer.Linker.SendPacket(LinkPacketHelpers.AdminCommand(userobj.name,
                        cobj.tUser.Name, cobj.cmdText.Split(new String[] { " " }, StringSplitOptions.None)[0], cobj.args));
                    return;
                }
            }
            
            ServerEvents.OnCommand(userobj, cobj.cmdText, cobj.tUser, cobj.args);
        }

        private static void ProcessAvatar(UserObject userobj, AresPacketReader packet)
        {
            if (!ServerEvents.OnAvatarReceived(userobj))
                return;

            byte[] new_avatar = packet.ReadBytes();

            if (userobj.orgAvatar != null)
                if (userobj.orgAvatar.SequenceEqual<byte>(new_avatar))
                    return; // it's a duplicate

            userobj.avatar = new_avatar.Length <= 1 ? null : new_avatar;
            userobj.orgAvatar = userobj.avatar;

            if (userobj.avatar != null)
            {
                if (userobj.cloaked)
                    UserPool.BroadcastToEqualOrHigher(userobj, ChatPacketHelpers.Avatar(userobj));
                else
                    UserPool.BroadcastToVroom(userobj.vroom, ChatPacketHelpers.Avatar(userobj));

                if (CallistoServer.Linker.IsLinking && userobj.vroom == 0)
                    CallistoServer.Linker.SendPacket(ChatPacketHelpers.Avatar(userobj));
            }
            else // blank
            {
                if (userobj.cloaked)
                    UserPool.BroadcastToEqualOrHigher(userobj, ChatPacketHelpers.AvatarCleared(userobj));
                else
                    UserPool.BroadcastToVroom(userobj.vroom, ChatPacketHelpers.AvatarCleared(userobj));

                if (CallistoServer.Linker.IsLinking && userobj.vroom == 0)
                    CallistoServer.Linker.SendPacket(ChatPacketHelpers.AvatarCleared(userobj));
            }
        }

        private static void ProcessAddSharedItem(UserObject userobj, AresPacketReader packet)
        {
            if (userobj.files.Count > 12000) // max file count is 12000
            {
                ServerEvents.OnFlood(userobj, ProtoMessage.MSG_CHAT_CLIENT_ADDSHARE);
                throw new Exception("too many share file items");
            }

            SharedFileObject item = new SharedFileObject();
            item.type = (MimeMessage)packet.ReadByte();
            item.size = packet.ReadInt32();
            ushort add_share_search_size = packet.ReadInt16();
            byte[] tmp = packet.ReadBytes((int)add_share_search_size);
            item.search = Encoding.UTF8.GetString(tmp).ToUpper();
            item.content = packet.ReadBytes();
            item.name = String.Empty;
            userobj.files.Add(item);

            if (item.type == MimeMessage.ARES_MIME_AUDIOOTHER1 ||
                item.type == MimeMessage.ARES_MIME_AUDIOOTHER2 ||
                item.type == MimeMessage.ARES_MIME_MP3)
                return;

            item.name = Helpers.FileItemFormatName(item.content, (byte)item.type, item.size);

            if (item.name != null)
                ServerEvents.OnFileReceived(userobj, item.name, item.type);
            else item.name = String.Empty;
        }

        private static void ProcessRemoveSharedItem(UserObject userobj, AresPacketReader packet)
        {
            uint size = packet.ReadInt32();
            userobj.files.RemoveAll(s => s.size == size);
        }

        private static void ProcessSearchRequest(UserObject userobj, AresPacketReader packet)
        {
            ushort id = packet.ReadInt16();
            packet.SkipByte();
            byte org_type = packet.ReadByte();
            ushort search_len = packet.ReadInt16();
            byte[] tmp = packet.ReadBytes(search_len);
            String search_words = Encoding.UTF8.GetString(tmp).ToUpper();
            List<byte> compressed_packets = new List<byte>();

            foreach (UserObject target in UserPool.Pool)
            {
                byte type = org_type;

                if (target.Connected)
                {
                    if (target.logged_in)
                    {
                        if (target.browsable)
                        {
                            SharedFileObject[] files;

                            if (type == 0 || type == 255)
                                files = target.files.ToArray();
                            else
                                files = target.files.FindAll(s => s.type == (MimeMessage)(type == 8 ? 0 : type)).ToArray();

                            if (files.Length == 0)
                                continue;

                            for (int i = 0; i < files.Length; i++)
                            {
                                if (files[i].search.Contains(search_words))
                                {
                                    byte[] p = ChatPacketHelpers.SearchHit(id, target, files[i]);

                                    if (userobj.supports_compression)
                                    {
                                        compressed_packets.AddRange(p);

                                        if (compressed_packets.Count > 800)
                                        {
                                            p = ChatPacketHelpers.ClientCompressed(compressed_packets.ToArray());
                                            compressed_packets.Clear();
                                            userobj.SendPacket(p);
                                        }
                                    }
                                    else { userobj.SendPacket(p); }
                                }
                            }
                        }
                    }
                }
            }

            if (userobj.supports_compression)
            {
                if (compressed_packets.Count > 0)
                {
                    byte[] p = ChatPacketHelpers.ClientCompressed(compressed_packets.ToArray());
                    compressed_packets.Clear();
                    userobj.SendPacket(p);
                }
            }

            userobj.SendPacket(ChatPacketHelpers.EndOfSearch(id));
        }

        private static void ProcessBrowseRequest(UserObject userobj, AresPacketReader packet)
        {
            ushort id = packet.ReadInt16();
            byte type = packet.ReadByte();
            String name = packet.ReadString();
            UserObject target = UserPool.FindUserByName(name);
            List<byte> compressed_packets = new List<byte>();

            if (target == null)
            {
                if (CallistoServer.Linker.IsLinking && userobj.vroom == 0)
                {
                    LinkedUserObject ltarget = CallistoServer.Linker.linked_users.Find(l => l.name == name);

                    if (ltarget != null)
                    {
                        if (type == 0) type = 255;
                        if (type == 8) type = 0;
                        CallistoServer.Linker.SendPacket(LinkPacketHelpers.BrowseRequest(userobj, ltarget, id, type));
                        return;
                    }
                }

                userobj.SendPacket(ChatPacketHelpers.BrowseError(id));
                return;
            }

            if (target.files.Count == 0)
            {
                userobj.SendPacket(ChatPacketHelpers.BrowseError(id));
                return;
            }

            SharedFileObject[] files;

            if (type == 0 || type == 255)
                files = target.files.ToArray();
            else
                files = target.files.FindAll(s => s.type == (MimeMessage)(type == 8 ? 0 : type)).ToArray();

            userobj.SendPacket(ChatPacketHelpers.StartOfBrowse(id, (ushort)files.Length));

            for (int i = 0; i < files.Length; i++)
            {
                byte[] p = ChatPacketHelpers.BrowseItem(id, files[i]);

                if (userobj.supports_compression)
                {
                    compressed_packets.AddRange(p);

                    if (compressed_packets.Count > 800)
                    {
                        p = ChatPacketHelpers.ClientCompressed(compressed_packets.ToArray());
                        compressed_packets.Clear();
                        userobj.SendPacket(p);
                    }
                }
                else { userobj.SendPacket(p); }
            }

            if (userobj.supports_compression)
            {
                if (compressed_packets.Count > 0)
                {
                    byte[] p = ChatPacketHelpers.ClientCompressed(compressed_packets.ToArray());
                    compressed_packets.Clear();
                    userobj.SendPacket(p);
                }
            }

            userobj.SendPacket(ChatPacketHelpers.EndOfBrowse(id));
        }

        private static void ProcessDirectChatPush(UserObject userobj, AresPacketReader packet)
        {
            if (packet.Remaining() < 4) return;

            String wantedNick = packet.ReadString();

            if (Encoding.UTF8.GetByteCount(wantedNick) < 2)
            {
                userobj.SendPacket(new byte[] { 1, 0, (byte)ProtoMessage.MSG_CHAT_CLIENT_DIRCHATPUSH, 4 });
                return;
            }

            if (packet.Remaining() != 16) // missing random string
            {
                userobj.SendPacket(new byte[] { 1, 0, (byte)ProtoMessage.MSG_CHAT_CLIENT_DIRCHATPUSH, 3 });
                return;
            }

            byte[] text_synch = packet.ReadBytes();

            UserObject wantedUser = UserPool.FindUserByName(wantedNick);

            if (wantedUser == null) // user not found
            {
                userobj.SendPacket(new byte[] { 1, 0, (byte)ProtoMessage.MSG_CHAT_CLIENT_DIRCHATPUSH, 1 });
                return;
            }

            if (wantedUser.ignores.Contains(userobj.name)) // user is ignoring you
            {
                userobj.SendPacket(new byte[] { 1, 0, (byte)ProtoMessage.MSG_CHAT_CLIENT_DIRCHATPUSH, 2 });
                return;
            }

            userobj.SendPacket(new byte[] { 1, 0, (byte)ProtoMessage.MSG_CHAT_CLIENT_DIRCHATPUSH, 0 });
            wantedUser.SendPacket(ChatPacketHelpers.DirectChatPush(userobj, text_synch));
        }

        private static void ProcessIgnoreRequest(UserObject userobj, AresPacketReader packet)
        {
            byte type = packet.ReadByte();
            String name = packet.ReadString();
            IUser target = UserPool.FindUserByName(name);

            if (target == null)
                target = WebPool.FindUserByName(name);

            if (target == null)
                if (CallistoServer.Linker.IsLinking && userobj.vroom == 0)
                    target = CallistoServer.Linker.linked_users.Find(l => l.name == name);

            if (target != null)
            {
                if (type == 0) // unignore
                {
                    if (userobj.ignores.Contains(target.Name))
                        userobj.ignores.Remove(target.Name);

                    userobj.SendPacket(ChatPacketHelpers.NoSuch(target.Name + " is unignored"));
                }
                else
                {
                    if (userobj.level < target.Level || target.IsLocalHost)
                    {
                        userobj.SendPacket(ChatPacketHelpers.NoSuch("You are not allowed to ignore this user"));
                    }
                    else
                    {
                        if (!userobj.ignores.Contains(target.Name))
                            userobj.ignores.Add(target.Name);

                        userobj.SendPacket(ChatPacketHelpers.NoSuch(target.Name + " is ignored"));
                    }
                }
            }
        }

        private static void ProcessCustomData(UserObject userobj, AresPacketReader packet)
        {
            String ident = packet.ReadString();
            String name = packet.ReadString();
            UserObject target = UserPool.FindUserByName(name);

            if (target == null)
            {
                if (CallistoServer.Linker.IsLinking && userobj.vroom == 0)
                {
                    LinkedUserObject ltarget = CallistoServer.Linker.linked_users.Find(l => l.name == name);

                    if (ltarget != null)
                        CallistoServer.Linker.SendPacket(LinkPacketHelpers.RelayToUser(ltarget.name, ChatPacketHelpers.CustomData(userobj, ident, packet.ReadBytes())));
                }
            }
            else { target.SendPacket(ChatPacketHelpers.CustomData(userobj, ident, packet.ReadBytes())); }
        }

        private static void ProcessCustomDataAll(UserObject userobj, AresPacketReader packet)
        {
            String ident = packet.ReadString();
            byte[] buf = packet.ReadBytes();

            if (ident.StartsWith("cb0t_scribble_"))
                if (!Settings.allow_room_scribbles)
                    return;

            byte[] data = ChatPacketHelpers.CustomData(userobj, ident, buf);

            UserPool.BroadcastCustomData(userobj, data);

            if (CallistoServer.Linker.IsLinking && userobj.vroom == 0)
                CallistoServer.Linker.SendPacket(LinkPacketHelpers.Broadcast(data));
        }

        private static void ProcessCustomDataAddTags(UserObject userobj, AresPacketReader packet)
        {
            while (packet.Remaining() > 0)
                userobj.custom_tags.Add(packet.ReadString());

            if (CallistoServer.Linker.IsLinking && userobj.vroom == 0)
                CallistoServer.Linker.SendPacket(LinkPacketHelpers.SetCustomTags(userobj));
        }

        private static void ProcessCustomDataRemTags(UserObject userobj, AresPacketReader packet)
        {
            while (packet.Remaining() > 0)
                userobj.custom_tags.RemoveAll(x => x == packet.ReadString());

            if (CallistoServer.Linker.IsLinking && userobj.vroom == 0)
                CallistoServer.Linker.SendPacket(LinkPacketHelpers.SetCustomTags(userobj));
        }

        private static void ProcessCustomFont(UserObject userobj, AresPacketReader packet)
        {
            bool clear_font = false;

            if (packet.Remaining() == 2)
            {
                userobj.font = new AresFont();
                clear_font = true;
            }
            else
            {
                userobj.font = new AresFont(packet.ReadByte(), packet.ReadString());

                if (packet.Remaining() >= 2)
                {
                    userobj.font.NameColor = packet.ReadByte();
                    userobj.font.TextColor = packet.ReadByte();
                }

                if (packet.Remaining() == 6)
                {
                    userobj.font.NameColorEx = packet.ReadBytes(3);
                    userobj.font.TextColorEx = packet.ReadBytes(3);
                }
            }

            UserPool.BroadcastToVroom(userobj.vroom, ChatPacketHelpers.CustomFont(userobj));
            WebPool.BroadcastAresFont(userobj);

            if (clear_font)
                userobj.font = null;
        }

        private static void ProcessVCFirst(UserObject userobj, AresPacketReader packet)
        {
            if (Settings.enable_vc)
            {
                byte[] buffer = ChatPacketHelpers.VoiceChatFirst(userobj.name, packet.ReadBytes());
                UserPool.BroadcastVoiceClip(userobj, buffer);
            }
        }

        private static void ProcessVCFirstTo(UserObject userobj, AresPacketReader packet)
        {
            if (Settings.enable_vc)
            {
                String target_name = packet.ReadString();
                byte[] buffer = ChatPacketHelpers.VoiceChatFirstTo(userobj.name, packet.ReadBytes());
                UserObject target_obj = UserPool.FindUserByName(target_name);

                if (target_obj != null)
                    if (!target_obj.voice_chat_ignores.Contains(userobj.name))
                        if (target_obj.voice_private_chat)
                            target_obj.SendPacket(buffer);
                        else userobj.SendPacket(ChatPacketHelpers.VoiceChatNoPrivate(target_name));
                    else userobj.SendPacket(ChatPacketHelpers.VoiceChatIgnored(target_name));
                else userobj.SendPacket(ChatPacketHelpers.OfflineUser(target_name));
            }
        }

        private static void ProcessVCChunk(UserObject userobj, AresPacketReader packet)
        {
            if (Settings.enable_vc)
            {
                byte[] buffer = ChatPacketHelpers.VoiceChatChunk(userobj.name, packet.ReadBytes());
                UserPool.BroadcastVoiceClip(userobj, buffer);
            }
        }

        private static void ProcessVCChunkTo(UserObject userobj, AresPacketReader packet)
        {
            if (Settings.enable_vc)
            {
                String target_name = packet.ReadString();
                byte[] buffer = ChatPacketHelpers.VoiceChatChunkTo(userobj.name, packet.ReadBytes());
                UserObject target_obj = UserPool.FindUserByName(target_name);

                if (target_obj != null)
                    if (target_obj.voice_private_chat)
                        if (!target_obj.voice_chat_ignores.Contains(userobj.name))
                            target_obj.SendPacket(buffer);
            }
        }

        private static void ProcessVCIgnore(UserObject userobj, AresPacketReader packet)
        {
            if (Settings.enable_vc)
            {
                String target_name = packet.ReadString();

                if (userobj.voice_chat_ignores.Contains(target_name))
                {
                    userobj.SendPacket(ChatPacketHelpers.NoSuch(target_name + " is voice chat unignored"));
                    userobj.voice_chat_ignores.RemoveAll(x => x == target_name);
                }
                else
                {
                    userobj.SendPacket(ChatPacketHelpers.NoSuch(target_name + " is voice chat ignored"));
                    userobj.voice_chat_ignores.Add(target_name);
                }
            }
        }

        private static void ProcessSupportsCustomEmotes(UserObject userobj)
        {
            if (!Settings.allow_custom_emotes)
                return;

            if (!userobj.supports_custom_emotes)
            {
                userobj.supports_custom_emotes = true;
                userobj.custom_emotes.Clear();
                UserPool.SendCustomEmotes(userobj);
            }
        }

        private static void ProcessCustomEmoteUpload(UserObject userobj, AresPacketReader packet)
        {
            if (!Settings.allow_custom_emotes)
                return;

            ProcessSupportsCustomEmotes(userobj);

            CustomEmoteItem c = new CustomEmoteItem();
            c.Shortcut = packet.ReadString();
            c.Size = packet.ReadByte();
            c.Image = packet.ReadBytes();

            if (userobj.cloaked)
                UserPool.BroadcastToEqualOrHigherCustomEmotes(userobj, ChatPacketHelpers.CustomEmoteItem(userobj, c));
            else
                UserPool.BroadcastToVroomCustomEmotes(userobj.vroom, ChatPacketHelpers.CustomEmoteItem(userobj, c));

            userobj.custom_emotes.Add(c);

            if (userobj.custom_emotes.Count > 16)
                throw new Exception("exceeded custom emoticon maximum");
        }

        private static void ProcessCustomEmoteDelete(UserObject userobj, AresPacketReader packet)
        {
            if (!Settings.allow_custom_emotes)
                return;

            String text = packet.ReadString();
            userobj.custom_emotes.RemoveAll(x => x.Shortcut == text);
            UserPool.BroadcastToVroomCustomEmotes(userobj.vroom, ChatPacketHelpers.CustomEmoteDelete(userobj, text));
        }

    }
}
