﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml;
using System.Net;

namespace callisto
{
    class UdpServerPool
    {
        private static Random rnd;
        public static List<UdpServerObject> Pool;
        public static UdpSortMethod SortMethod;

        public static void AddServer(UdpServerObject server)
        {
            if (server.ip.Equals(Settings.ServerIP)) // don't want our own room on the list since that will be sent anyway
                return;

            UdpServerObject sobj = Pool.Find(s => s.ip.Equals(server.ip));

            if (sobj != null) // maybe room port has changed?
            {
                sobj.port = server.port;
                return;
            }

            if (Pool.Count > 4000)
                PurgeList();

            Pool.Add(server);
        }

        public static void ExpireList()
        {
            uint time = Helpers.UnixTime;
            SortMethod = UdpSortMethod.SortMostRecentFirst;
            Pool.Sort();
            Pool.RemoveAll(s => s.try_count > 4 && (time - s.last_sent_add_IPS) > 60 && (time - s.last_connect) > 3600);
        }

        public static void PurgeList()
        {
            SortMethod = UdpSortMethod.SortMostRecentFirst;
            Pool.Sort();

            uint time = Helpers.UnixTime;
            int never_counted = Pool.FindAll(s => s.last_sent_add_IPS == 0).Count;
            int successful = Pool.FindAll(s => (s.last_connect + 900) > time).Count;

            Pool.Reverse();

            if (never_counted <= 600 && successful >= 50)
            {
                int target = Pool.FindIndex(s => s.last_sent_add_IPS == 0);

                if (target > -1)
                {
                    Pool.RemoveAt(target);
                    return;
                }
            }

            SortMethod = UdpSortMethod.SortMostAckedFirst;
            Pool.Sort();
            Pool.RemoveAt(Pool.Count - 1);
        }

        public static UdpServerObject[] GetServers(int max_servers)
        {
            ShuffleList();

            uint time = Helpers.UnixTime;
            List<UdpServerObject> results = Pool.FindAll(s => s.ack_count > 0 && (s.last_connect + 900) > time);

            if (results.Count > max_servers)
                results = results.GetRange(0, max_servers);

            return results.ToArray();
        }

        public static UdpServerObject[] GetServers(IPAddress target_ip, int max_servers)
        {
            ShuffleList();

            uint time = Helpers.UnixTime;
            List<UdpServerObject> results = Pool.FindAll(s => !s.ip.Equals(target_ip) && s.ack_count > 0 && (s.last_connect + 900) > time);

            if (results.Count > max_servers)
                results = results.GetRange(0, max_servers);

            return results.ToArray();
        }

        public static void Load()
        {
            Pool = new List<UdpServerObject>();
            SortMethod = UdpSortMethod.SortMostRecentFirst;

            try
            {
                FileStream f = new FileStream(Settings.Path + "udp.xml", FileMode.Open);
                XmlReader xml = XmlReader.Create(new StreamReader(f));

                xml.MoveToContent();
                xml.ReadSubtree().ReadToFollowing("servers");

                while (xml.ReadToFollowing("server"))
                {
                    UdpServerObject c = new UdpServerObject();

                    xml.ReadSubtree().ReadToFollowing("ip");
                    c.ip = new IPAddress(Convert.FromBase64String(xml.ReadElementContentAsString()));
                    xml.ReadToFollowing("port");
                    c.port = BitConverter.ToUInt16(Convert.FromBase64String(xml.ReadElementContentAsString()), 0);
                    xml.ReadToFollowing("ack");
                    c.ack_count = (int)BitConverter.ToUInt16(Convert.FromBase64String(xml.ReadElementContentAsString()), 0);

                    Pool.Add(c);
                }

                xml.Close();
                f.Flush();
                f.Close();
            }
            catch { }

            if (Pool.Count < 5)
                Download();
        }

        public static void Save()
        {
            int counter = 0;

            try
            {
                XmlWriterSettings appearance = new XmlWriterSettings();
                appearance.Indent = true;
                XmlWriter xml = XmlWriter.Create(Settings.Path + "udp.xml", appearance);

                xml.WriteStartDocument();
                xml.WriteStartElement("servers");

                SortMethod = UdpSortMethod.SortMostRecentFirst;
                Pool.Sort();
                
                int time = (int)Helpers.UnixTime;

                foreach (UdpServerObject c in Pool)
                {
                    if (c.ack_count > 0)
                    {
                        if ((time - c.last_connect) < 1800)
                        {
                            if (c.ack_count > 65000)
                                c.ack_count = 40000;

                            xml.WriteStartElement("server");
                            xml.WriteElementString("ip", Convert.ToBase64String(c.ip.GetAddressBytes()));
                            xml.WriteElementString("port", Convert.ToBase64String(BitConverter.GetBytes(c.port)));
                            xml.WriteElementString("ack", Convert.ToBase64String(BitConverter.GetBytes((ushort)c.ack_count)));
                            xml.WriteEndElement();
                            counter++;

                            if (counter == 100)
                                break;
                        }
                    }
                }

                xml.WriteEndElement();
                xml.WriteEndDocument();
                xml.Flush();
                xml.Close();
            }
            catch { }

            UdpStats.Reset();

            if (counter == 0)
            {
                Pool.Clear();
                Download();
            }
        }

        public static void ShuffleList()
        {
            if (rnd == null)
                rnd = new Random();

            int n = Pool.Count;

            while (n > 1)
            {
                n--;
                int k = rnd.Next(n + 1);
                UdpServerObject value = Pool[k];
                Pool[k] = Pool[n];
                Pool[n] = value;
            }
        }

        private static void Download()
        {
            try
            {
                List<byte> received_data = new List<byte>(File.ReadAllBytes(AppDomain.CurrentDomain.BaseDirectory + "gcache.dat"));

                while (received_data.Count > 6)
                {
                    IPAddress ip = new IPAddress(received_data.GetRange(0, 4).ToArray());
                    received_data.RemoveRange(0, 4);
                    ushort port = BitConverter.ToUInt16(received_data.ToArray(), 0);
                    received_data.RemoveRange(0, 2);
                    Pool.Add(new UdpServerObject(ip, port));
                }
            }
            catch { }
        }

    }
}
